/*
*
*	Example implementation for "Analytic Spectral Integration of Birefringence-Induced Iridescence"
*		by Shlomi Steinberg
*
* This code is licensed under the MIT license (MIT) (http://opensource.org/licenses/MIT)
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <vector>
#include <random>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <ctime>
#include <sstream>
#include <filesystem>
#include <optional>

#define GLM_FORCE_RADIANS
#define GLM_FORCE_DEPTH_ZERO_TO_ONE
#include <glm/glm.hpp>
#include <glm/gtc/matrix_transform.hpp>
#include <glm/gtc/matrix_inverse.hpp>

#include <vulkan/vulkan.h>
#include "base/vulkanexamplebase.h"
#include "base/VulkanBuffer.hpp"
#include "base/VulkanTexture.hpp"
#include "base/VulkanModel.hpp"

// color-util expects M_PI to be defined, for reasons
#define M_PI glm::pi<double>()
#include <color-util/XYZ_to_Lab.hpp>
#include <color-util/CIEDE2000.hpp>
#undef M_PI

#ifdef _linux
#include <png.h>
#elif defined(_WIN32)
#include <FreeImage.h>
#else
#error Unkown platform
#endif


#define VERTEX_BUFFER_BIND_ID 0
#define ENABLE_VALIDATION false

#ifdef _linux
inline bool write_2d(const char* file_name, const uint8_t *image_data, int components, int width, int height)
{
	FILE *fp = fopen(file_name, "wb");
	if (!fp) {
		std::cerr << file_name << " can't be opened for writing" << std::endl;
		return false;
	}

	png_structp png = png_create_write_struct(PNG_LIBPNG_VER_STRING, nullptr, nullptr, nullptr);
	if (!png) {
		std::cerr << file_name << " png_create_write_struct failed" << std::endl;
		fclose(fp);
		return false;
	}

	png_infop info = png_create_info_struct(png);
	if (!info) {
		std::cerr << file_name << " png_create_info_struct failed" << std::endl;
		fclose(fp);
		return false;
	}

	if (setjmp(png_jmpbuf(png))) {
		std::cerr << file_name << " png_jmpbuf failed" << std::endl;
		fclose(fp);
		return false;
	}

	png_byte ** const row_pointers = reinterpret_cast<png_byte **>(malloc(height * sizeof(png_byte *)));
	if (row_pointers == nullptr) {
		std::cerr << file_name << " could not allocate memory for PNG row pointers" << std::endl;
		fclose(fp);
		return false;
	}

	// set the individual row_pointers to point at the correct offsets of image_data
	// To maintain compatibility png_write_image requests a non-const double pointer, hack the const away...
	for (int i = 0; i < height; i++)
		row_pointers[i] = const_cast<png_byte*>(image_data + i * width * components);

	png_init_io(png, fp);

	int color_type;
	switch (components) {
	case 1: color_type = PNG_COLOR_TYPE_GRAY; break;
	case 3: color_type = PNG_COLOR_TYPE_RGB; break;
	case 4: color_type = PNG_COLOR_TYPE_RGBA; break;
	default: {
		std::cerr << "Invalid component count" << std::endl;
		return false;
	}
	}
	png_set_IHDR(png,
		info,
		width, height,
		8,
		color_type,
		PNG_INTERLACE_NONE,
		PNG_COMPRESSION_TYPE_DEFAULT,
		PNG_FILTER_TYPE_DEFAULT);
	png_write_info(png, info);

	png_write_image(png, row_pointers);
	png_write_end(png, nullptr);

	free(row_pointers);
	fclose(fp);

	return true;
}
#elif defined(_WIN32)
inline bool write_2d(const char* file_name, const uint8_t *image_data, int components, int width, int height)
{
	FIBITMAP *img = FreeImage_ConvertFromRawBits(const_cast<BYTE*>(image_data), width, height, components * width, components * 8, FI_RGBA_RED_MASK, FI_RGBA_GREEN_MASK, FI_RGBA_BLUE_MASK, true);
	const auto ret = FreeImage_Save(FIF_PNG, img, file_name);
	FreeImage_Unload(img);

	return ret;
}
#endif

class VulkanExample : public VulkanExampleBase
{
public:
	struct {
		VkPipelineVertexInputStateCreateInfo inputState;
		std::vector<VkVertexInputBindingDescription> bindingDescriptions;
		std::vector<VkVertexInputAttributeDescription> attributeDescriptions;
	} vertices;

	// Vertex layout for the models
	vks::VertexLayout vertexLayout = vks::VertexLayout({
		vks::VERTEX_COMPONENT_POSITION,
		vks::VERTEX_COMPONENT_NORMAL,
		vks::VERTEX_COMPONENT_UV,
		vks::VERTEX_COMPONENT_COLOR,
	});

	struct {
		vks::Model skybox;
		vks::Model object;
	} models;

	struct {
		std::vector<vks::TextureCubeMap> environmentCubes;
		vks::Texture2D spectral_preintegrated_lut;
	} textures;

	struct {
		vks::Buffer object, skybox, params;
	} uniformBuffers;

	struct UBOVS {
		glm::mat4 projection;
		glm::mat4 modelView;
		glm::mat4 invModelView;
		glm::vec4 cameraPosition;
		glm::vec2 framebuffer_size{};
		int32_t envprobe = 1;
		float polarization = .5f;
		float noise_amount = .0f;
	} uboVS;
	struct Material {
		glm::vec4 albedo = glm::vec4{ 1.f };
		glm::vec4 A = glm::vec4{ 0.05,0,0.9,0 };
		glm::vec4 mu_o = glm::vec4{ 0,0,1, 0.01f };
		glm::vec4 mu_e = glm::vec4{ 0,0,1, 0.01f };
		float eta_o = 1.39f;
		float eta_e = 1.4f;
		float _unused0, _unused1;
	};
	struct UBOParams {
		Material material;
		float exposure = 1.0f;
		float d = 0.3;
	} uboParams;

	int32_t view_mode = 0;
	int32_t m_idx = 0;
	int32_t m = 0;
	int32_t inject_noise = 0;
	int32_t capture_mode = 0;

	struct {
		VkPipeline render;
		VkPipeline skybox;
		VkPipeline composition;
	} pipelines;
	struct {
		VkDescriptorSet object;
		VkDescriptorSet skybox;
		VkDescriptorSet composition;
	} descriptorSets;
	struct {
		VkPipelineLayout render;
		VkPipelineLayout composition;
	} pipelineLayouts;
	struct {
		VkDescriptorSetLayout render;
		VkDescriptorSetLayout composition;
	} descriptorSetLayouts;

	struct FrameBufferAttachment {
		VkImage image;
		VkDeviceMemory mem;
		VkImageView view;
		VkFormat format;
		void destroy(VkDevice device)
		{
			vkDestroyImageView(device, view, nullptr);
			vkDestroyImage(device, image, nullptr);
			vkFreeMemory(device, mem, nullptr);
		}
	};
	struct FrameBuffer {
		int32_t width, height;
		VkFramebuffer frameBuffer;
		FrameBufferAttachment color;
		FrameBufferAttachment depth;
		VkRenderPass renderPass;
		VkSampler sampler;
		VkCommandBuffer cmdBuffer = VK_NULL_HANDLE;
		VkSemaphore semaphore = VK_NULL_HANDLE;
	} offscreen;

	struct delta_capture_t {
		uint32_t width, height;
		int32_t m, m_idx;
		std::string name;
		std::unique_ptr<glm::vec3[]> src;
	};
	std::optional<delta_capture_t> delta_capture;

	bool beginDeltaCapture(std::string name, int32_t capture_m, int32_t capture_m_idx) {
		delta_capture_t capture = { width, height, m, m_idx, std::move(name) };
		capture.src = captureBackbuffer();
		if (!capture.src)
			return false;

		delta_capture = std::move(capture);
		m = capture_m;
		m_idx = capture_m_idx;

		return true;
	}
	bool endDeltaCapture() {
		// End capture
		const auto dst = captureBackbuffer();
		const auto src = std::move(delta_capture.value().src);
		const auto capture_m = m;
		m = delta_capture.value().m;
		m_idx = delta_capture.value().m_idx;
		auto name = std::move(delta_capture.value().name);
		capture_mode = 0;

		if (!dst || !src || width != delta_capture.value().width || height != delta_capture.value().height) {
			delta_capture = {};
			return false;
		}

		const auto RGBtoXYZ = glm::inverse(glm::mat3( 2.0413690, -0.5649464, -0.3446944,
													 -0.9692660,  1.8760108,  0.0415560,
													  0.0134474, -0.1183897,  1.0154096));

		// Create diff image
		const auto pixels = width*height;
		auto ciede2000 = std::make_unique<double[]>(pixels);
		double greatest_diff = .0;
		for (int x=0;x<pixels;++x) {
			const auto srcXYZ = src[x] * RGBtoXYZ;
			const auto dstXYZ = dst[x] * RGBtoXYZ;
			const auto srclab = colorutil::convert_XYZ_to_Lab(colorutil::XYZ{ srcXYZ.x, srcXYZ.y, srcXYZ.z });
			const auto dstlab = colorutil::convert_XYZ_to_Lab(colorutil::XYZ{ dstXYZ.x, dstXYZ.y, dstXYZ.z });

			ciede2000[x] = colorutil::calculate_CIEDE2000(srclab, dstlab);
			greatest_diff = std::max(greatest_diff, ciede2000[x]);
		}
		std::cout << "Greatest CIEDE2000 difference: " << std::to_string(greatest_diff) << std::endl;

		// Read colormap
		std::vector<glm::vec3> colormap;
		{
			std::string colormap_path = getAssetPath() + "magma.csv";
			std::ifstream ifs(colormap_path);
			if (!ifs) {
				std::cout << "'" << colormap_path << "' not found!" << std::endl;
				throw std::runtime_error("missing data");
			}
			std::string dummy;
			std::getline(ifs, dummy);
			std::vector<double> data;
			colormap.reserve(300);
			while ((data = readCSVline(ifs)).size())
				colormap.emplace_back(glm::vec3{ static_cast<float>(data[0]), static_cast<float>(data[1]), static_cast<float>(data[2]) });
		}

		// Create bitmaps
		auto pdst = std::make_unique<uint8_t[]>(width*height*3);
		auto pdiff = std::make_unique<uint8_t[]>(width*height*3);
		for (uint32_t y = 0; y < height; y++) {
			for (uint32_t x = 0; x < width; x++) {
				const auto idx = y*width + x;
				
				auto packed_dst = glm::packUnorm4x8(glm::vec4{ dst[idx], .0f });
				// Colour code
				const float err = glm::pow(ciede2000[idx], 2.3);
				const auto cmidxf = err * 3.1;
				const auto cm0 = colormap[std::min(static_cast<size_t>(cmidxf),   colormap.size()-1)];
				const auto cm1 = colormap[std::min(static_cast<size_t>(cmidxf)+1, colormap.size()-1)];
				const auto cm = glm::mix(cm0, cm1, glm::fract(cmidxf));
				auto packed_diff = glm::packUnorm4x8(glm::vec4{ cm, .0f });

				pdst[(y*width+x)*3+2] = static_cast<uint8_t>((packed_dst>>0) &0xFF);
				pdst[(y*width+x)*3+1] = static_cast<uint8_t>((packed_dst>>8) &0xFF);
				pdst[(y*width+x)*3+0] = static_cast<uint8_t>((packed_dst>>16)&0xFF);
				pdiff[(y*width+x)*3+2] = static_cast<uint8_t>((packed_diff>>0) &0xFF);
				pdiff[(y*width+x)*3+1] = static_cast<uint8_t>((packed_diff>>8) &0xFF);
				pdiff[(y*width+x)*3+0] = static_cast<uint8_t>((packed_diff>>16)&0xFF);
			}
		}

		// Writeout PNGs
		if (write_2d((name+"_dst" + std::to_string(capture_m) + ".png").data(), pdst.get(), 3, width, height) &&
			write_2d((name+"_diff" + std::to_string(capture_m) + ".png").data(), pdiff.get(), 3, width, height))
			std::cout << "capture saved to '" << name << "_X.png'" << std::endl;

		// End capture
		delta_capture = {};
		return true;
	}

	VulkanExample() : VulkanExampleBase(ENABLE_VALIDATION)
	{
		title = "Analytic Spectral Integration of Birefringence-Induced Iridescence";
		settings.overlay = true;

		camera.type = Camera::CameraType::lookat;
		camera.setPerspective(60.0f, (float)width / (float)height, 0.1f, 256.0f);
		camera.setRotation({ -3.75f, 150.0f, 0.0f });
		camera.setPosition({ 0.f, 0.f, -.6f });
	}

	~VulkanExample()
	{
		// Clean up used Vulkan resources 
		vkDestroyPipeline(device, pipelines.skybox, nullptr);
		vkDestroyPipeline(device, pipelines.render, nullptr);
		vkDestroyPipeline(device, pipelines.composition, nullptr);
		vkDestroyPipelineLayout(device, pipelineLayouts.render, nullptr);
		vkDestroyPipelineLayout(device, pipelineLayouts.composition, nullptr);
		vkDestroyDescriptorSetLayout(device, descriptorSetLayouts.render, nullptr);
		vkDestroyDescriptorSetLayout(device, descriptorSetLayouts.composition, nullptr);

		models.object.destroy();
		models.skybox.destroy();		

		uniformBuffers.object.destroy();
		uniformBuffers.skybox.destroy();
		uniformBuffers.params.destroy();

		vkDestroySemaphore(device, offscreen.semaphore, nullptr);
		vkDestroyRenderPass(device, offscreen.renderPass, nullptr);
		vkDestroyFramebuffer(device, offscreen.frameBuffer, nullptr);
		vkDestroySampler(device, offscreen.sampler, nullptr);

		offscreen.depth.destroy(device);
		offscreen.color.destroy(device);
	}

	std::vector<double> readCSVline(std::istream& str) {
		std::vector<double> result;
		std::string l;
		std::getline(str,l);

		std::stringstream ss(l);
		std::string s;
		while (std::getline(ss, s, ','))
			result.push_back(std::stod(s));
		return result;

	}

	void resizeAssets()
	{}

	void loadAssets()
	{
		models.skybox.loadFromFile(getAssetPath() + "cube.obj", vertexLayout, 1.0f, vulkanDevice, queue);

		models.object.loadFromFile(getAssetPath() + "cube.obj", vertexLayout, 0.075f, vulkanDevice, queue);
		//models.object.loadFromFile(getAssetPath() + "chinesedragon.dae", vertexLayout, 0.075f, vulkanDevice, queue);
		//models.object.loadFromFile(getAssetPath() + "sibenik.dae", vertexLayout, 0.021f, vulkanDevice, queue);

		textures.environmentCubes.resize(3);
		textures.environmentCubes[0].loadFromFile(getAssetPath() + "hdr/gcanyon_cube.ktx", VK_FORMAT_R16G16B16A16_SFLOAT, vulkanDevice, queue);
		textures.environmentCubes[1].loadFromFile(getAssetPath() + "hdr/uffizi_cube.ktx", VK_FORMAT_R16G16B16A16_SFLOAT, vulkanDevice, queue);
		textures.environmentCubes[2].loadFromFile(getAssetPath() + "hdr/pisa_cube.ktx", VK_FORMAT_R16G16B16A16_SFLOAT, vulkanDevice, queue);

		// Load precomputed spectral integrals data
		std::string lut_path = getAssetPath() + "spectral_preintegrated.csv";
		{
			std::ifstream ifs(lut_path);
			if (!ifs) {
				std::cout << "'" << lut_path << "' not found!" << std::endl;
				throw std::runtime_error("missing data");
			}

			constexpr int N = 1024*8;
			std::array<float, N*4> texture;

			std::string dummy;
			std::getline(ifs, dummy);
			std::vector<double> data;
			std::size_t lines_read = 0;
			while ((data = readCSVline(ifs)).size()) {
				assert(lines_read < N && data.size() == 4);

				texture[lines_read*4 + 0] = static_cast<float>(data[1]);
				texture[lines_read*4 + 1] = static_cast<float>(data[2]);
				texture[lines_read*4 + 2] = static_cast<float>(data[3]);
				texture[lines_read*4 + 3] = 1.f;

				++lines_read;
			}

			std::cout << "Read " << std::to_string(lines_read) << " from '" << lut_path << "'." << std::endl;

			assert(lines_read == N);

			textures.spectral_preintegrated_lut.fromBuffer(&texture[0], sizeof(texture), VK_FORMAT_R32G32B32A32_SFLOAT, N, 1,vulkanDevice, queue, 
														   VK_FILTER_LINEAR, VK_IMAGE_USAGE_SAMPLED_BIT, VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL, VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE, VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE);
		}
	}

	void reBuildCommandBuffers()
	{
		if (!checkCommandBuffers())
		{
			destroyCommandBuffers();
			createCommandBuffers();
		}
		buildCommandBuffers();
	}

	void buildCommandBuffers()
	{		
		{
			if (offscreen.cmdBuffer == VK_NULL_HANDLE)
			{
				offscreen.cmdBuffer = VulkanExampleBase::createCommandBuffer(VK_COMMAND_BUFFER_LEVEL_PRIMARY, false);
			}
			if (offscreen.semaphore == VK_NULL_HANDLE)
			{
				VkSemaphoreCreateInfo semaphoreCreateInfo = vks::initializers::semaphoreCreateInfo();
				VK_CHECK_RESULT(vkCreateSemaphore(device, &semaphoreCreateInfo, nullptr, &offscreen.semaphore));
			}

			VkCommandBufferBeginInfo cmdBufInfo = vks::initializers::commandBufferBeginInfo();

			// Clear values for all attachments written in the fragment sahder
			std::array<VkClearValue, 2> clearValues;
			clearValues[0].color = defaultClearColor;
			clearValues[1].depthStencil = { 1.0f, 0 };

			VkRenderPassBeginInfo renderPassBeginInfo = vks::initializers::renderPassBeginInfo();
			renderPassBeginInfo.renderPass = offscreen.renderPass;
			renderPassBeginInfo.framebuffer = offscreen.frameBuffer;
			renderPassBeginInfo.renderArea.extent.width = offscreen.width;
			renderPassBeginInfo.renderArea.extent.height = offscreen.height;
			renderPassBeginInfo.clearValueCount = clearValues.size();
			renderPassBeginInfo.pClearValues = clearValues.data();

			VK_CHECK_RESULT(vkBeginCommandBuffer(offscreen.cmdBuffer, &cmdBufInfo));

			vkCmdBeginRenderPass(offscreen.cmdBuffer, &renderPassBeginInfo, VK_SUBPASS_CONTENTS_INLINE);

			VkViewport viewport = vks::initializers::viewport((float)offscreen.width, (float)offscreen.height, 0.0f, 1.0f);
			vkCmdSetViewport(offscreen.cmdBuffer, 0, 1, &viewport);
			VkRect2D scissor = vks::initializers::rect2D(offscreen.width, offscreen.height, 0, 0);
			vkCmdSetScissor(offscreen.cmdBuffer, 0, 1, &scissor);
			VkDeviceSize offsets[1] = { 0 };

			vkCmdBindDescriptorSets(offscreen.cmdBuffer, VK_PIPELINE_BIND_POINT_GRAPHICS, pipelineLayouts.render, 0, 1, &descriptorSets.skybox, 0, NULL);
			vkCmdBindPipeline(offscreen.cmdBuffer, VK_PIPELINE_BIND_POINT_GRAPHICS, pipelines.skybox);
			vkCmdBindVertexBuffers(offscreen.cmdBuffer, 0, 1, &models.skybox.vertices.buffer, offsets);
			vkCmdBindIndexBuffer(offscreen.cmdBuffer, models.skybox.indices.buffer, 0, VK_INDEX_TYPE_UINT32);
			vkCmdDrawIndexed(offscreen.cmdBuffer, models.skybox.indexCount, 1, 0, 0, 0);

			vkCmdNextSubpass(offscreen.cmdBuffer, VK_SUBPASS_CONTENTS_INLINE);

			vkCmdBindDescriptorSets(offscreen.cmdBuffer, VK_PIPELINE_BIND_POINT_GRAPHICS, pipelineLayouts.render, 0, 1, &descriptorSets.object, 0, NULL);
			vkCmdBindPipeline(offscreen.cmdBuffer, VK_PIPELINE_BIND_POINT_GRAPHICS, pipelines.render);
			vkCmdBindVertexBuffers(offscreen.cmdBuffer, VERTEX_BUFFER_BIND_ID, 1, &models.object.vertices.buffer, offsets);
			vkCmdBindIndexBuffer(offscreen.cmdBuffer, models.object.indices.buffer, 0, VK_INDEX_TYPE_UINT32);
			vkCmdDrawIndexed(offscreen.cmdBuffer, models.object.indexCount, 1, 0, 0, 0);
			
			vkCmdEndRenderPass(offscreen.cmdBuffer);

			VK_CHECK_RESULT(vkEndCommandBuffer(offscreen.cmdBuffer));
		}

		VkCommandBufferBeginInfo cmdBufInfo = vks::initializers::commandBufferBeginInfo();

		VkClearValue clearValues[2];
		clearValues[0].color = { { 0.0f, 0.0f, 0.0f, 0.0f } };
		clearValues[1].depthStencil = { 1.0f, 0 };

		VkRenderPassBeginInfo renderPassBeginInfo = vks::initializers::renderPassBeginInfo();
		renderPassBeginInfo.renderPass = renderPass;
		renderPassBeginInfo.renderArea.offset.x = 0;
		renderPassBeginInfo.renderArea.offset.y = 0;
		renderPassBeginInfo.renderArea.extent.width = width;
		renderPassBeginInfo.renderArea.extent.height = height;
		renderPassBeginInfo.clearValueCount = 2;
		renderPassBeginInfo.pClearValues = clearValues;

		for (int32_t i = 0; i < drawCmdBuffers.size(); ++i)
		{
			// Set target frame buffer
			renderPassBeginInfo.framebuffer = frameBuffers[i];

			VK_CHECK_RESULT(vkBeginCommandBuffer(drawCmdBuffers[i], &cmdBufInfo));

			const VkDeviceSize offsets[1] = { 0 };
			vkCmdBeginRenderPass(drawCmdBuffers[i], &renderPassBeginInfo, VK_SUBPASS_CONTENTS_INLINE);
			VkViewport viewport = vks::initializers::viewport((float)width, (float)height, 0.0f, 1.0f);
			vkCmdSetViewport(drawCmdBuffers[i], 0, 1, &viewport);
			VkRect2D scissor = vks::initializers::rect2D(width, height,	0, 0);
			vkCmdSetScissor(drawCmdBuffers[i], 0, 1, &scissor);
			
			vkCmdBindDescriptorSets(drawCmdBuffers[i], VK_PIPELINE_BIND_POINT_GRAPHICS, pipelineLayouts.composition, 0, 1, &descriptorSets.composition, 0, NULL);
			vkCmdBindPipeline(drawCmdBuffers[i], VK_PIPELINE_BIND_POINT_GRAPHICS, pipelines.composition);
			vkCmdDraw(drawCmdBuffers[i], 3, 1, 0, 0);

			vkCmdEndRenderPass(drawCmdBuffers[i]);

			VK_CHECK_RESULT(vkEndCommandBuffer(drawCmdBuffers[i]));
		}
	}

	bool saveScreenshot(const char *filename) {
		const auto data = captureBackbuffer();
		if (!data)
			return false;

		auto p = std::make_unique<uint8_t[]>(width*height*3);
		for (uint32_t y = 0; y < height; y++) {
			for (uint32_t x = 0; x < width; x++) {
				const glm::vec3 c0 = data[y*width+x];
				glm::vec4 c1 = glm::vec4{ c0.x, c0.y, c0.z, .0f };
				auto packed = glm::packUnorm4x8(c1);
				p[(y*width+x)*3+0] = static_cast<uint8_t>((packed>>0) &0xFF);
				p[(y*width+x)*3+1] = static_cast<uint8_t>((packed>>8) &0xFF);
				p[(y*width+x)*3+2] = static_cast<uint8_t>((packed>>16)&0xFF);
			}
		}
		
		if (write_2d(filename, p.get(), 3, width, height))
			std::cout << "Screenshot saved to '" << filename << "'" << std::endl;

		return true;
	}

	std::unique_ptr<glm::vec3[]> captureBackbuffer() {
		// Source for the copy is the last rendered swapchain image
		VkImage srcImage = offscreen.color.image;
	
		// Create the linear tiled destination image to copy to and to read the memory from
		VkImageCreateInfo imageCreateCI(vks::initializers::imageCreateInfo());
		imageCreateCI.imageType = VK_IMAGE_TYPE_2D;
		// Note that vkCmdBlitImage (if supported) will also do format conversions if the swapchain color format would differ
		imageCreateCI.format = VK_FORMAT_R32G32B32A32_SFLOAT;
		imageCreateCI.extent.width = width;
		imageCreateCI.extent.height = height;
		imageCreateCI.extent.depth = 1;
		imageCreateCI.arrayLayers = 1;
		imageCreateCI.mipLevels = 1;
		imageCreateCI.initialLayout = VK_IMAGE_LAYOUT_UNDEFINED;
		imageCreateCI.samples = VK_SAMPLE_COUNT_1_BIT;
		imageCreateCI.tiling = VK_IMAGE_TILING_LINEAR;
		imageCreateCI.usage = VK_IMAGE_USAGE_TRANSFER_DST_BIT;
		// Create the image
		VkImage dstImage;
		VK_CHECK_RESULT(vkCreateImage(device, &imageCreateCI, nullptr, &dstImage));
		// Create memory to back up the image
		VkMemoryRequirements memRequirements;
		VkMemoryAllocateInfo memAllocInfo(vks::initializers::memoryAllocateInfo());
		VkDeviceMemory dstImageMemory;
		vkGetImageMemoryRequirements(device, dstImage, &memRequirements);
		memAllocInfo.allocationSize = memRequirements.size;
		// Memory must be host visible to copy from
		memAllocInfo.memoryTypeIndex = vulkanDevice->getMemoryType(memRequirements.memoryTypeBits, VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT);
		VK_CHECK_RESULT(vkAllocateMemory(device, &memAllocInfo, nullptr, &dstImageMemory));
		VK_CHECK_RESULT(vkBindImageMemory(device, dstImage, dstImageMemory, 0));

		// Do the actual blit from the swapchain image to our host visible destination image
		VkCommandBuffer copyCmd = vulkanDevice->createCommandBuffer(VK_COMMAND_BUFFER_LEVEL_PRIMARY, true);

		VkImageMemoryBarrier imageMemoryBarrier = vks::initializers::imageMemoryBarrier();
		
		// Transition destination image to transfer destination layout
		vks::tools::insertImageMemoryBarrier(
			copyCmd,
			dstImage,
			0,
			VK_ACCESS_TRANSFER_WRITE_BIT,
			VK_IMAGE_LAYOUT_UNDEFINED,
			VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,
			VK_PIPELINE_STAGE_TRANSFER_BIT,
			VK_PIPELINE_STAGE_TRANSFER_BIT,
			VkImageSubresourceRange{ VK_IMAGE_ASPECT_COLOR_BIT, 0, 1, 0, 1 });

		// Transition swapchain image from present to transfer source layout
		vks::tools::insertImageMemoryBarrier(
			copyCmd,
			srcImage,
			VK_ACCESS_MEMORY_READ_BIT,
			VK_ACCESS_TRANSFER_READ_BIT,
			VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL,
			VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL,
			VK_PIPELINE_STAGE_TRANSFER_BIT,
			VK_PIPELINE_STAGE_TRANSFER_BIT,
			VkImageSubresourceRange{ VK_IMAGE_ASPECT_COLOR_BIT, 0, 1, 0, 1 });

		VkImageCopy imageCopyRegion{};
		imageCopyRegion.srcSubresource.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
		imageCopyRegion.srcSubresource.layerCount = 1;
		imageCopyRegion.dstSubresource.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
		imageCopyRegion.dstSubresource.layerCount = 1;
		imageCopyRegion.extent.width = width;
		imageCopyRegion.extent.height = height;
		imageCopyRegion.extent.depth = 1;

		// Issue the copy command
		vkCmdCopyImage(
			copyCmd,
			srcImage, VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL,
			dstImage, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,
			1,
			&imageCopyRegion);

		// Transition destination image to general layout, which is the required layout for mapping the image memory later on
		vks::tools::insertImageMemoryBarrier(
			copyCmd,
			dstImage,
			VK_ACCESS_TRANSFER_WRITE_BIT,
			VK_ACCESS_MEMORY_READ_BIT,
			VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,
			VK_IMAGE_LAYOUT_GENERAL,
			VK_PIPELINE_STAGE_TRANSFER_BIT,
			VK_PIPELINE_STAGE_TRANSFER_BIT,
			VkImageSubresourceRange{ VK_IMAGE_ASPECT_COLOR_BIT, 0, 1, 0, 1 });

		// Transition back the swap chain image after the blit is done
		vks::tools::insertImageMemoryBarrier(
			copyCmd,
			srcImage,
			VK_ACCESS_TRANSFER_READ_BIT,
			VK_ACCESS_MEMORY_READ_BIT,
			VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL,
			VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL,
			VK_PIPELINE_STAGE_TRANSFER_BIT,
			VK_PIPELINE_STAGE_TRANSFER_BIT,
			VkImageSubresourceRange{ VK_IMAGE_ASPECT_COLOR_BIT, 0, 1, 0, 1 });

		vulkanDevice->flushCommandBuffer(copyCmd, queue);

		// Get layout of the image (including row pitch)
		VkImageSubresource subResource { VK_IMAGE_ASPECT_COLOR_BIT, 0, 0 };
		VkSubresourceLayout subResourceLayout;
		vkGetImageSubresourceLayout(device, dstImage, &subResource, &subResourceLayout);

		// Map image memory so we can start copying from it
		const char* data;
		vkMapMemory(device, dstImageMemory, 0, VK_WHOLE_SIZE, 0, (void**)&data);
		data += subResourceLayout.offset;

		auto p = std::make_unique<glm::vec3[]>(width*height);
		for (uint32_t y = 0; y < height; y++) {
			const auto *row = (const glm::vec4*)data;
			for (uint32_t x = 0; x < width; x++) {
				const glm::vec3 c0 = glm::vec3(1.f) - glm::exp(-glm::vec3{ *row } * uboParams.exposure);
				const glm::vec4 c1 = glm::vec4(glm::pow(c0, glm::vec3(1.f / 2.2f)), 1.f);
				p[y*width+x] = glm::vec3{ c1.x, c1.y, c1.z };
				++row;
			}
			data += subResourceLayout.rowPitch;
		}

		// Clean up resources
		vkUnmapMemory(device, dstImageMemory);
		vkFreeMemory(device, dstImageMemory, nullptr);
		vkDestroyImage(device, dstImage, nullptr);

		return p;
	}

	void createAttachment(VkFormat format, VkImageUsageFlagBits usage, FrameBufferAttachment *attachment)
	{
		VkImageAspectFlags aspectMask = 0;
		VkImageLayout imageLayout;

		attachment->format = format;

		if (usage & VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT)
		{
			aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
			imageLayout = VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL;
		}
		if (usage & VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT)
		{
			aspectMask = VK_IMAGE_ASPECT_DEPTH_BIT | VK_IMAGE_ASPECT_STENCIL_BIT;
			imageLayout = VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL;
		}

		assert(aspectMask > 0);

		VkImageCreateInfo image = vks::initializers::imageCreateInfo();
		image.imageType = VK_IMAGE_TYPE_2D;
		image.format = format;
		image.extent.width = offscreen.width;
		image.extent.height = offscreen.height;
		image.extent.depth = 1;
		image.mipLevels = 1;
		image.arrayLayers = 1;
		image.samples = VK_SAMPLE_COUNT_1_BIT;
		image.tiling = VK_IMAGE_TILING_OPTIMAL;
		image.usage = usage | VK_IMAGE_USAGE_SAMPLED_BIT;

		VkMemoryAllocateInfo memAlloc = vks::initializers::memoryAllocateInfo();
		VkMemoryRequirements memReqs;

		VK_CHECK_RESULT(vkCreateImage(device, &image, nullptr, &attachment->image));
		vkGetImageMemoryRequirements(device, attachment->image, &memReqs);
		memAlloc.allocationSize = memReqs.size;
		memAlloc.memoryTypeIndex = vulkanDevice->getMemoryType(memReqs.memoryTypeBits, VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT);
		VK_CHECK_RESULT(vkAllocateMemory(device, &memAlloc, nullptr, &attachment->mem));
		VK_CHECK_RESULT(vkBindImageMemory(device, attachment->image, attachment->mem, 0));

		VkImageViewCreateInfo imageView = vks::initializers::imageViewCreateInfo();
		imageView.viewType = VK_IMAGE_VIEW_TYPE_2D;
		imageView.format = format;
		imageView.subresourceRange = {};
		imageView.subresourceRange.aspectMask = aspectMask;
		imageView.subresourceRange.baseMipLevel = 0;
		imageView.subresourceRange.levelCount = 1;
		imageView.subresourceRange.baseArrayLayer = 0;
		imageView.subresourceRange.layerCount = 1;
		imageView.image = attachment->image;
		VK_CHECK_RESULT(vkCreateImageView(device, &imageView, nullptr, &attachment->view));
	}

	void setupFrameBuffer() override
	{
		VulkanExampleBase::setupFrameBuffer();

		offscreen.width = width;
		offscreen.height = height;

		// Two floating point color buffers
		createAttachment(VK_FORMAT_R32G32B32A32_SFLOAT, static_cast<VkImageUsageFlagBits>(VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT | VK_IMAGE_USAGE_TRANSFER_SRC_BIT), &offscreen.color);
		// Depth attachment
		createAttachment(depthFormat, VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT, &offscreen.depth);

		// Set up separate renderpass with references to the color and depth attachments
		std::array<VkAttachmentDescription, 2> attachmentDescs = {};

		// Init attachment properties
		for (uint32_t i = 0; i < attachmentDescs.size(); ++i)
		{
			attachmentDescs[i].samples = VK_SAMPLE_COUNT_1_BIT;
			attachmentDescs[i].loadOp = VK_ATTACHMENT_LOAD_OP_DONT_CARE;
			attachmentDescs[i].storeOp = VK_ATTACHMENT_STORE_OP_STORE;
			attachmentDescs[i].stencilLoadOp = VK_ATTACHMENT_LOAD_OP_DONT_CARE;
			attachmentDescs[i].stencilStoreOp = VK_ATTACHMENT_STORE_OP_DONT_CARE;
			attachmentDescs[i].initialLayout = VK_IMAGE_LAYOUT_UNDEFINED;
		}
		attachmentDescs[0].finalLayout = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;
		attachmentDescs[0].format = offscreen.color.format;
		attachmentDescs[1].finalLayout = VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL;
		attachmentDescs[1].format = offscreen.depth.format;
		attachmentDescs[1].loadOp = VK_ATTACHMENT_LOAD_OP_CLEAR;
		attachmentDescs[1].storeOp = VK_ATTACHMENT_STORE_OP_DONT_CARE;

		std::vector<VkAttachmentReference> colorReferences;
		colorReferences.push_back({ 0, VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL });

		VkAttachmentReference depthReference = {};
		depthReference.attachment = 1;
		depthReference.layout = VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL;

		VkSubpassDescription subpass[2] = {};
		subpass[0].pipelineBindPoint = VK_PIPELINE_BIND_POINT_GRAPHICS;
		subpass[0].pColorAttachments = colorReferences.data();
		subpass[0].colorAttachmentCount = 1;
		subpass[0].pDepthStencilAttachment = &depthReference;
		subpass[1].pipelineBindPoint = VK_PIPELINE_BIND_POINT_GRAPHICS;
		subpass[1].pColorAttachments = colorReferences.data();
		subpass[1].colorAttachmentCount = colorReferences.size();
		subpass[1].pDepthStencilAttachment = &depthReference;

		// Use subpass dependencies for attachment layput transitions
		std::array<VkSubpassDependency, 3> dependencies;

		dependencies[0].srcSubpass = VK_SUBPASS_EXTERNAL;
		dependencies[0].dstSubpass = 0;
		dependencies[0].srcStageMask = VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT;
		dependencies[0].dstStageMask = VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT;
		dependencies[0].srcAccessMask = VK_ACCESS_MEMORY_READ_BIT;
		dependencies[0].dstAccessMask = VK_ACCESS_COLOR_ATTACHMENT_READ_BIT | VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT;
		dependencies[0].dependencyFlags = VK_DEPENDENCY_BY_REGION_BIT;

		dependencies[1].srcSubpass = 0;
		dependencies[1].dstSubpass = 1;
		dependencies[1].srcStageMask = VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT;
		dependencies[1].dstStageMask = VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT;
		dependencies[1].srcAccessMask = VK_ACCESS_COLOR_ATTACHMENT_READ_BIT | VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT;
		dependencies[1].dstAccessMask = VK_ACCESS_MEMORY_READ_BIT;
		dependencies[1].dependencyFlags = VK_DEPENDENCY_BY_REGION_BIT;

		dependencies[2].srcSubpass = 1;
		dependencies[2].dstSubpass = VK_SUBPASS_EXTERNAL;
		dependencies[2].srcStageMask = VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT;
		dependencies[2].dstStageMask = VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT;
		dependencies[2].srcAccessMask = VK_ACCESS_COLOR_ATTACHMENT_READ_BIT | VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT;
		dependencies[2].dstAccessMask = VK_ACCESS_MEMORY_READ_BIT;
		dependencies[2].dependencyFlags = VK_DEPENDENCY_BY_REGION_BIT;

		VkRenderPassCreateInfo renderPassInfo = {};
		renderPassInfo.sType = VK_STRUCTURE_TYPE_RENDER_PASS_CREATE_INFO;
		renderPassInfo.pAttachments = attachmentDescs.data();
		renderPassInfo.attachmentCount = static_cast<uint32_t>(attachmentDescs.size());
		renderPassInfo.subpassCount = 2;
		renderPassInfo.pSubpasses = subpass;
		renderPassInfo.dependencyCount = dependencies.size();
		renderPassInfo.pDependencies = dependencies.data();

		VK_CHECK_RESULT(vkCreateRenderPass(device, &renderPassInfo, nullptr, &offscreen.renderPass));

		std::array<VkImageView, 2> attachments;
		attachments[0] = offscreen.color.view;
		attachments[1] = offscreen.depth.view;

		VkFramebufferCreateInfo fbufCreateInfo = {};
		fbufCreateInfo.sType = VK_STRUCTURE_TYPE_FRAMEBUFFER_CREATE_INFO;
		fbufCreateInfo.pNext = NULL;
		fbufCreateInfo.renderPass = offscreen.renderPass;
		fbufCreateInfo.pAttachments = attachments.data();
		fbufCreateInfo.attachmentCount = static_cast<uint32_t>(attachments.size());
		fbufCreateInfo.width = offscreen.width;
		fbufCreateInfo.height = offscreen.height;
		fbufCreateInfo.layers = 1;
		VK_CHECK_RESULT(vkCreateFramebuffer(device, &fbufCreateInfo, nullptr, &offscreen.frameBuffer));

		// Create sampler to sample from the color attachments
		VkSamplerCreateInfo sampler = vks::initializers::samplerCreateInfo();
		sampler.magFilter = VK_FILTER_NEAREST;
		sampler.minFilter = VK_FILTER_NEAREST;
		sampler.mipmapMode = VK_SAMPLER_MIPMAP_MODE_LINEAR;
		sampler.addressModeU = VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE;
		sampler.addressModeV = sampler.addressModeU;
		sampler.addressModeW = sampler.addressModeU;
		sampler.mipLodBias = 0.0f;
		sampler.maxAnisotropy = 1.0f;
		sampler.minLod = 0.0f;
		sampler.maxLod = 1.0f;
		sampler.borderColor = VK_BORDER_COLOR_FLOAT_OPAQUE_WHITE;
		VK_CHECK_RESULT(vkCreateSampler(device, &sampler, nullptr, &offscreen.sampler));

		resizeAssets();
	}

	void prepareVertices()
	{
		// Binding description
		vertices.bindingDescriptions.resize(1);
		vertices.bindingDescriptions[0] =
			vks::initializers::vertexInputBindingDescription(
				VERTEX_BUFFER_BIND_ID,
				vertexLayout.stride(),
				VK_VERTEX_INPUT_RATE_VERTEX);

		// Attribute descriptions
		// Describes memory layout and shader positions
		vertices.attributeDescriptions.resize(4);
		// Location 0 : Position
		vertices.attributeDescriptions[0] =
			vks::initializers::vertexInputAttributeDescription(
				VERTEX_BUFFER_BIND_ID,
				0,
				VK_FORMAT_R32G32B32_SFLOAT,
				0);
		// Location 1 : Normal
		vertices.attributeDescriptions[1] =
			vks::initializers::vertexInputAttributeDescription(
				VERTEX_BUFFER_BIND_ID,
				1,
				VK_FORMAT_R32G32B32_SFLOAT,
				sizeof(float) * 3);
		// Location 2 : Texture coordinates
		vertices.attributeDescriptions[2] =
			vks::initializers::vertexInputAttributeDescription(
				VERTEX_BUFFER_BIND_ID,
				2,
				VK_FORMAT_R32G32_SFLOAT,
				sizeof(float) * 6);
		// Location 3 : Color
		vertices.attributeDescriptions[3] =
			vks::initializers::vertexInputAttributeDescription(
				VERTEX_BUFFER_BIND_ID,
				3,
				VK_FORMAT_R32G32B32_SFLOAT,
				sizeof(float) * 8);

		vertices.inputState = vks::initializers::pipelineVertexInputStateCreateInfo();
		vertices.inputState.vertexBindingDescriptionCount = vertices.bindingDescriptions.size();
		vertices.inputState.pVertexBindingDescriptions = vertices.bindingDescriptions.data();
		vertices.inputState.vertexAttributeDescriptionCount = vertices.attributeDescriptions.size();
		vertices.inputState.pVertexAttributeDescriptions = vertices.attributeDescriptions.data();
	}

	void setupDescriptorPool()
	{
		// Example uses one ubo and one image sampler
		std::vector<VkDescriptorPoolSize> poolSizes =
		{
			vks::initializers::descriptorPoolSize(VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, 8),
			vks::initializers::descriptorPoolSize(VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, 12)
		};

		VkDescriptorPoolCreateInfo descriptorPoolInfo =
			vks::initializers::descriptorPoolCreateInfo(
				poolSizes.size(),
				poolSizes.data(),
				3);

		VK_CHECK_RESULT(vkCreateDescriptorPool(device, &descriptorPoolInfo, nullptr, &descriptorPool));
	}

	void setupDescriptorSetLayout()
	{
		std::vector<VkDescriptorSetLayoutBinding> setLayoutBindings =
		{
			vks::initializers::descriptorSetLayoutBinding(VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, VK_SHADER_STAGE_VERTEX_BIT | VK_SHADER_STAGE_FRAGMENT_BIT, 0),
			vks::initializers::descriptorSetLayoutBinding(VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, VK_SHADER_STAGE_FRAGMENT_BIT, 1, textures.environmentCubes.size()),
			vks::initializers::descriptorSetLayoutBinding(VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, VK_SHADER_STAGE_VERTEX_BIT | VK_SHADER_STAGE_FRAGMENT_BIT, 2),
			vks::initializers::descriptorSetLayoutBinding(VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, VK_SHADER_STAGE_FRAGMENT_BIT, 3),
		};

		VkDescriptorSetLayoutCreateInfo descriptorLayout =
			vks::initializers::descriptorSetLayoutCreateInfo(
				setLayoutBindings.data(),
				setLayoutBindings.size());

		VK_CHECK_RESULT(vkCreateDescriptorSetLayout(device, &descriptorLayout, nullptr, &descriptorSetLayouts.render));

		VkPipelineLayoutCreateInfo pPipelineLayoutCreateInfo =
			vks::initializers::pipelineLayoutCreateInfo(
				&descriptorSetLayouts.render,
				1);
		pPipelineLayoutCreateInfo.pushConstantRangeCount = 0;

		VK_CHECK_RESULT(vkCreatePipelineLayout(device, &pPipelineLayoutCreateInfo, nullptr, &pipelineLayouts.render));

		{
			setLayoutBindings = {
				vks::initializers::descriptorSetLayoutBinding(VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, VK_SHADER_STAGE_FRAGMENT_BIT, 0),
				vks::initializers::descriptorSetLayoutBinding(VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, VK_SHADER_STAGE_FRAGMENT_BIT, 1)
			};

			descriptorLayout = vks::initializers::descriptorSetLayoutCreateInfo(setLayoutBindings.data(), static_cast<uint32_t>(setLayoutBindings.size()));
			VK_CHECK_RESULT(vkCreateDescriptorSetLayout(device, &descriptorLayout, nullptr, &descriptorSetLayouts.composition));

			pPipelineLayoutCreateInfo = vks::initializers::pipelineLayoutCreateInfo(&descriptorSetLayouts.composition, 1);
			pPipelineLayoutCreateInfo.pushConstantRangeCount = 0;
			VK_CHECK_RESULT(vkCreatePipelineLayout(device, &pPipelineLayoutCreateInfo, nullptr, &pipelineLayouts.composition));
		}
	}

	void setupDescriptorSet()
	{
		std::vector<VkDescriptorImageInfo> envcubemaps;
		for (const auto &c : textures.environmentCubes)
			envcubemaps.emplace_back(c.descriptor);

		VkDescriptorSetAllocateInfo allocInfo =
			vks::initializers::descriptorSetAllocateInfo(
				descriptorPool,
				&descriptorSetLayouts.render,
				1);

		VK_CHECK_RESULT(vkAllocateDescriptorSets(device, &allocInfo, &descriptorSets.object));

		std::vector<VkWriteDescriptorSet> writeDescriptorSets =
		{
			vks::initializers::writeDescriptorSet(descriptorSets.object, VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, 0, &uniformBuffers.object.descriptor),
			vks::initializers::writeDescriptorSet(descriptorSets.object, VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, 1, envcubemaps.data(), envcubemaps.size()),
			vks::initializers::writeDescriptorSet(descriptorSets.object, VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, 2, &uniformBuffers.params.descriptor),
			vks::initializers::writeDescriptorSet(descriptorSets.object, VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, 3, &textures.spectral_preintegrated_lut.descriptor),
		};

		vkUpdateDescriptorSets(device, writeDescriptorSets.size(), writeDescriptorSets.data(), 0, NULL);

		{
			// Sky box
			VK_CHECK_RESULT(vkAllocateDescriptorSets(device, &allocInfo, &descriptorSets.skybox));

			writeDescriptorSets = {
				vks::initializers::writeDescriptorSet(descriptorSets.skybox, VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, 0, &uniformBuffers.skybox.descriptor),
				vks::initializers::writeDescriptorSet(descriptorSets.skybox, VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, 2, &uniformBuffers.params.descriptor),
				vks::initializers::writeDescriptorSet(descriptorSets.skybox, VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, 1, envcubemaps.data(), envcubemaps.size())
			};
			vkUpdateDescriptorSets(device, static_cast<uint32_t>(writeDescriptorSets.size()), writeDescriptorSets.data(), 0, NULL);
		}

		{
			allocInfo =	vks::initializers::descriptorSetAllocateInfo(descriptorPool, &descriptorSetLayouts.composition, 1);
			VK_CHECK_RESULT(vkAllocateDescriptorSets(device, &allocInfo, &descriptorSets.composition));

			std::vector<VkDescriptorImageInfo> colorDescriptors = {
				vks::initializers::descriptorImageInfo(offscreen.sampler, offscreen.color.view, VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL),
			};

			writeDescriptorSets = {
				vks::initializers::writeDescriptorSet(descriptorSets.composition, VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, 0, &colorDescriptors[0]),
				vks::initializers::writeDescriptorSet(descriptorSets.composition, VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, 1, &uniformBuffers.params.descriptor)
			};
			vkUpdateDescriptorSets(device, static_cast<uint32_t>(writeDescriptorSets.size()), writeDescriptorSets.data(), 0, NULL);
		}
	}

	void preparePipelines()
	{
		VkPipelineInputAssemblyStateCreateInfo inputAssemblyState =
			vks::initializers::pipelineInputAssemblyStateCreateInfo(
				VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST,
				0,
				VK_FALSE);

		VkPipelineRasterizationStateCreateInfo rasterizationState =
			vks::initializers::pipelineRasterizationStateCreateInfo(
				VK_POLYGON_MODE_FILL,
				VK_CULL_MODE_BACK_BIT,
				VK_FRONT_FACE_CLOCKWISE,
				0);

		VkPipelineColorBlendAttachmentState blendAttachmentState[1] = {
			vks::initializers::pipelineColorBlendAttachmentState(0xf, VK_FALSE),
		};

		VkPipelineColorBlendStateCreateInfo colorBlendState =
			vks::initializers::pipelineColorBlendStateCreateInfo(
				1,
				blendAttachmentState);

		VkPipelineDepthStencilStateCreateInfo depthStencilState =
			vks::initializers::pipelineDepthStencilStateCreateInfo(
				VK_TRUE,
				VK_TRUE,
				VK_COMPARE_OP_LESS_OR_EQUAL);

		VkPipelineViewportStateCreateInfo viewportState =
			vks::initializers::pipelineViewportStateCreateInfo(1, 1, 0);

		VkPipelineMultisampleStateCreateInfo multisampleState =
			vks::initializers::pipelineMultisampleStateCreateInfo(
				VK_SAMPLE_COUNT_1_BIT,
				0);

		std::vector<VkDynamicState> dynamicStateEnables = {
			VK_DYNAMIC_STATE_VIEWPORT,
			VK_DYNAMIC_STATE_SCISSOR
		};
		VkPipelineDynamicStateCreateInfo dynamicState =
			vks::initializers::pipelineDynamicStateCreateInfo(
				dynamicStateEnables.data(),
				dynamicStateEnables.size(),
				0);
				
		constexpr int specializationConstants = 3;
		std::array<VkSpecializationMapEntry, specializationConstants> specializationMapEntries;
		specializationMapEntries[0].constantID = 0;
		specializationMapEntries[0].size = 4;
		specializationMapEntries[0].offset = 0;
		specializationMapEntries[1].constantID = 1;
		specializationMapEntries[1].size = 4;
		specializationMapEntries[1].offset = 4;
		specializationMapEntries[2].constantID = 2;
		specializationMapEntries[2].size = 4;
		specializationMapEntries[2].offset = 8;
		int32_t constants[specializationConstants] = {};
		constants[0] = view_mode;
		constants[1] = m;
		constants[2] = inject_noise;
		VkSpecializationInfo specializationInfo{};
		specializationInfo.dataSize = sizeof(constants);
		specializationInfo.mapEntryCount = static_cast<uint32_t>(specializationMapEntries.size());
		specializationInfo.pMapEntries = specializationMapEntries.data();
		specializationInfo.pData = constants;

		// Load shaders
		std::array<VkPipelineShaderStageCreateInfo, 2> shaderStages;
		shaderStages[0] = loadShader(getAssetPath() + "../shaders/render/sem.vert.spv", VK_SHADER_STAGE_VERTEX_BIT);
		shaderStages[1] = loadShader(getAssetPath() + "../shaders/render/sem.frag.spv", VK_SHADER_STAGE_FRAGMENT_BIT);
		shaderStages[1].pSpecializationInfo = &specializationInfo;

		VkGraphicsPipelineCreateInfo pipelineCreateInfo =
			vks::initializers::pipelineCreateInfo(
				pipelineLayouts.render,
				offscreen.renderPass,
				0);
		pipelineCreateInfo.pVertexInputState = &vertices.inputState;
		pipelineCreateInfo.pInputAssemblyState = &inputAssemblyState;
		pipelineCreateInfo.pRasterizationState = &rasterizationState;
		pipelineCreateInfo.pColorBlendState = &colorBlendState;
		pipelineCreateInfo.pMultisampleState = &multisampleState;
		pipelineCreateInfo.pViewportState = &viewportState;
		pipelineCreateInfo.pDepthStencilState = &depthStencilState;
		pipelineCreateInfo.pDynamicState = &dynamicState;
		pipelineCreateInfo.stageCount = shaderStages.size();
		pipelineCreateInfo.pStages = shaderStages.data();
		pipelineCreateInfo.subpass = 1;

		VK_CHECK_RESULT(vkCreateGraphicsPipelines(device, pipelineCache, 1, &pipelineCreateInfo, nullptr, &pipelines.render));

		{
			// Skybox

			std::vector<VkVertexInputBindingDescription> vertexInputBindings = {
				vks::initializers::vertexInputBindingDescription(0, vertexLayout.stride(), VK_VERTEX_INPUT_RATE_VERTEX),
			};
			// Attribute descriptions
			std::vector<VkVertexInputAttributeDescription> vertexInputAttributes = {
				vks::initializers::vertexInputAttributeDescription(0, 0, VK_FORMAT_R32G32B32_SFLOAT, 0),					// Position
				vks::initializers::vertexInputAttributeDescription(0, 1, VK_FORMAT_R32G32B32_SFLOAT, sizeof(float) * 3),	// Normal
				vks::initializers::vertexInputAttributeDescription(0, 2, VK_FORMAT_R32G32_SFLOAT, sizeof(float) * 6),		// UV
			};
			colorBlendState = vks::initializers::pipelineColorBlendStateCreateInfo(1, blendAttachmentState);

			VkPipelineVertexInputStateCreateInfo vertexInputState = vks::initializers::pipelineVertexInputStateCreateInfo();
			vertexInputState.vertexBindingDescriptionCount = static_cast<uint32_t>(vertexInputBindings.size());
			vertexInputState.pVertexBindingDescriptions = vertexInputBindings.data();
			vertexInputState.vertexAttributeDescriptionCount = static_cast<uint32_t>(vertexInputAttributes.size());
			vertexInputState.pVertexAttributeDescriptions = vertexInputAttributes.data();

			VkPipelineRasterizationStateCreateInfo rasterizationState =
				vks::initializers::pipelineRasterizationStateCreateInfo(VK_POLYGON_MODE_FILL, VK_CULL_MODE_NONE, VK_FRONT_FACE_COUNTER_CLOCKWISE);

			pipelineCreateInfo.pRasterizationState = &rasterizationState;
			pipelineCreateInfo.pVertexInputState = &vertexInputState;
			pipelineCreateInfo.subpass = 0;

			shaderStages[0] = loadShader(getAssetPath() +  "../shaders/base/skybox.vert.spv", VK_SHADER_STAGE_VERTEX_BIT);
			shaderStages[1] = loadShader(getAssetPath() +  "../shaders/base/skybox.frag.spv", VK_SHADER_STAGE_FRAGMENT_BIT);
			shaderStages[1].pSpecializationInfo = &specializationInfo;

			VK_CHECK_RESULT(vkCreateGraphicsPipelines(device, pipelineCache, 1, &pipelineCreateInfo, nullptr, &pipelines.skybox));
		}

		{
			VkPipelineVertexInputStateCreateInfo emptyInputState = vks::initializers::pipelineVertexInputStateCreateInfo();
			pipelineCreateInfo.pVertexInputState = &emptyInputState;

			VkPipelineRasterizationStateCreateInfo rasterizationState =
				vks::initializers::pipelineRasterizationStateCreateInfo(VK_POLYGON_MODE_FILL, VK_CULL_MODE_FRONT_BIT, VK_FRONT_FACE_COUNTER_CLOCKWISE);
			pipelineCreateInfo.pRasterizationState = &rasterizationState;

			// Final fullscreen composition pass pipeline
			shaderStages[0] = loadShader(getAssetPath() + "../shaders/hdr/composition.vert.spv", VK_SHADER_STAGE_VERTEX_BIT);
			shaderStages[1] = loadShader(getAssetPath() + "../shaders/hdr/composition.frag.spv", VK_SHADER_STAGE_FRAGMENT_BIT);
			pipelineCreateInfo.layout = pipelineLayouts.composition;
			pipelineCreateInfo.renderPass = renderPass;
			VK_CHECK_RESULT(vkCreateGraphicsPipelines(device, pipelineCache, 1, &pipelineCreateInfo, nullptr, &pipelines.composition));
		}
	}

	void prepareUniformBuffers()
	{
		// Vertex shader uniform buffer block
		VK_CHECK_RESULT(vulkanDevice->createBuffer(
			VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT,
			VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT,
			&uniformBuffers.object,
			sizeof(uboVS)));

		// Skybox vertex shader uniform buffer
		VK_CHECK_RESULT(vulkanDevice->createBuffer(
			VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT,
			VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT,
			&uniformBuffers.skybox,
			sizeof(uboVS)));

		// Tonemapping params
		VK_CHECK_RESULT(vulkanDevice->createBuffer(
			VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT,
			VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT,
			&uniformBuffers.params,
			sizeof(uboParams)));

		// Map persistent
		VK_CHECK_RESULT(uniformBuffers.object.map());
		VK_CHECK_RESULT(uniformBuffers.skybox.map());
		VK_CHECK_RESULT(uniformBuffers.params.map());

		updateUniformBuffers();
	}

	void updateUniformBuffers()
	{
		uboVS.cameraPosition = glm::vec4(camera.position, 1.0f);
		uboVS.projection = camera.matrices.perspective;
		uboVS.modelView = camera.matrices.view;
		uboVS.invModelView = glm::inverseTranspose(uboVS.modelView);
		uboVS.framebuffer_size = glm::vec2{ static_cast<float>(this->width), static_cast<float>(this->height) };
		memcpy(uniformBuffers.object.mapped, &uboVS, sizeof(uboVS));
		
		// Skybox
		uboVS.modelView = glm::mat4(glm::mat3(camera.matrices.view));
		memcpy(uniformBuffers.skybox.mapped, &uboVS, sizeof(uboVS));
	}
	void updateParams()
	{
		if (!inject_noise)
			uboParams.material.A = glm::vec4{ glm::normalize(glm::vec3{ uboParams.material.A.x, uboParams.material.A.y, uboParams.material.A.z }), .0f };
		memcpy(uniformBuffers.params.mapped, &uboParams, sizeof(uboParams));
	}

	void draw()
	{
		VulkanExampleBase::prepareFrame();

		submitInfo.pWaitSemaphores = &semaphores.presentComplete;
		submitInfo.pSignalSemaphores = &offscreen.semaphore;
		submitInfo.commandBufferCount = 1;
		submitInfo.pCommandBuffers = &offscreen.cmdBuffer;
		VK_CHECK_RESULT(vkQueueSubmit(queue, 1, &submitInfo, VK_NULL_HANDLE));

		submitInfo.pWaitSemaphores = &offscreen.semaphore;
		submitInfo.pSignalSemaphores = &semaphores.renderComplete;
		submitInfo.commandBufferCount = 1;
		submitInfo.pCommandBuffers = &drawCmdBuffers[currentBuffer];
		VK_CHECK_RESULT(vkQueueSubmit(queue, 1, &submitInfo, VK_NULL_HANDLE));

		VulkanExampleBase::submitFrame();
	}

	void prepare()
	{
		VulkanExampleBase::prepare();

		loadAssets();
		
		prepareVertices();
		prepareUniformBuffers();
		setupDescriptorSetLayout();
		preparePipelines();
		setupDescriptorPool();
		setupDescriptorSet();
		buildCommandBuffers();

		updateParams();
		
		prepared = true;
	}

	virtual void render() override
	{
		if (!prepared)
			return;
		
		draw();

		if (delta_capture) {
			// End rendering and capture
			endDeltaCapture();
			
			// Rebuild pipelines
			vkDeviceWaitIdle(device);
			preparePipelines();
			reBuildCommandBuffers();
		}
	}


	virtual void viewChanged() override
	{
		updateUniformBuffers();
	}
	virtual void windowResized() override 
	{
		setupDescriptorSetLayout();
		preparePipelines();
		setupDescriptorPool();
		setupDescriptorSet();
		buildCommandBuffers();
		
		updateUniformBuffers();
	}

	virtual void OnUpdateUIOverlay(vks::UIOverlay *overlay) override
	{
		bool should_update_materials = false;
		bool should_rebuild_pipelines = false;
		if (overlay->header("Settings")) {
			if (overlay->inputInt("Environment", &uboVS.envprobe)) {
				uboVS.envprobe = glm::clamp<int>(uboVS.envprobe, 0, textures.environmentCubes.size()-1);
				updateUniformBuffers();
			} 
			ImGui::SameLine();
			if (overlay->inputFloat("Exposure", &uboParams.exposure, 0.025f, 3))
				updateParams();
			if (overlay->sliderFloat("Polarization (0: s-polarized, 1: p-polarized)", &uboVS.polarization, .001f, 0.999f))
				updateUniformBuffers();
			if (overlay->comboBox("Polarization filter", &view_mode, { std::string("No filter"), std::string("s-polarization filter"), std::string("p-polarization filter") }))
				should_rebuild_pipelines = true;
			if (overlay->comboBox("mode", &m_idx, { std::string("Analytic integration (m=1)"), 
													std::string("3-samples numeric"), 
													std::string("8-samples numeric"), 
													std::string("16-samples numeric"), 
													std::string("64-samples numeric"), 
													std::string("256-samples numeric"), 
													std::string("1024-samples numeric"), })) {
				switch (m_idx) {
					default: m = 0;    break;
					case 1:  m = 3;    break;
					case 2:  m = 8;    break;
					case 3:  m = 16;   break;
					case 4:  m = 64;   break;
					case 5:  m = 256;  break;
					case 6:  m = 1024; break;
				}
				should_rebuild_pipelines = true;
			}
		}

		if (overlay->header("Material")) {
			should_update_materials |= overlay->sliderFloat("thickness (mm)", &uboParams.d, .0f, 10.f);
			// should_update_materials |= overlay->sliderFloat("##R", &uboParams.material.albedo.r, .0f, 1.f); ImGui::SameLine();
			// should_update_materials |= overlay->sliderFloat("##G", &uboParams.material.albedo.g, .0f, 1.f); ImGui::SameLine();
			// should_update_materials |= overlay->sliderFloat(u8"α##B", &uboParams.material.albedo.b, .0f, 1.f);
			if (overlay->checkBox("Use perlin noise", &inject_noise)) {
				uboParams.material.A = inject_noise ? glm::vec4{ .1f,.1f,.0f,.0f } : glm::vec4{ .0f,1.f,.0f,.0f };
				updateUniformBuffers();
				should_rebuild_pipelines = true;
			}
			if (!inject_noise) {
				if (overlay->sliderFloat(u8"##Ax", &uboParams.material.A.x,  -1.f, 1.f)) {
					const glm::vec2 yz = uboParams.material.A.y!=0 || uboParams.material.A.z!=0 ?
						glm::normalize(glm::vec2{ uboParams.material.A.y,uboParams.material.A.z }) * glm::sqrt(glm::max(.0f, 1.f - uboParams.material.A.x*uboParams.material.A.x)) :
						glm::vec2{ 0,0 };
					uboParams.material.A.y = yz.x;
					uboParams.material.A.z = yz.y;
					should_update_materials = true;
				}
				ImGui::SameLine();
				if (overlay->sliderFloat(u8"##Ay", &uboParams.material.A.y, -1.f, 1.f)) {
					const glm::vec2 xz = uboParams.material.A.x!=0 || uboParams.material.A.z!=0 ?
						glm::normalize(glm::vec2{ uboParams.material.A.x,uboParams.material.A.z }) * glm::sqrt(glm::max(.0f, 1.f - uboParams.material.A.y*uboParams.material.A.y)):
						glm::vec2{ 0,0 };
					uboParams.material.A.x = xz.x;
					uboParams.material.A.z = xz.y;
					should_update_materials = true;
				}
				ImGui::SameLine();
				if (overlay->sliderFloat(u8"A##Az", &uboParams.material.A.z, -1.f, 1.f)) {
					const glm::vec2 xy = uboParams.material.A.x!=0 || uboParams.material.A.y!=0 ?
						glm::normalize(glm::vec2{ uboParams.material.A.x,uboParams.material.A.y }) * glm::sqrt(glm::max(.0f, 1.f - uboParams.material.A.z*uboParams.material.A.z)):
						glm::vec2{ 0,0 };
					uboParams.material.A.x = xy.x;
					uboParams.material.A.y = xy.y;
					should_update_materials = true;
				}
			}
			else {
				if (overlay->sliderFloat(u8"##Nx", &uboParams.material.A.x,  .0f, 1.f))
					should_update_materials = true;
				ImGui::SameLine();
				if (overlay->sliderFloat(u8"Noise##Ny", &uboParams.material.A.y, 0.f, 1.f))
					should_update_materials = true;
			}
			should_update_materials |= overlay->sliderFloat(u8"η_o##IORo", &uboParams.material.eta_o, 1.1f, 2.f); ImGui::SameLine();
			should_update_materials |= overlay->sliderFloat(u8"η_e##IORe", &uboParams.material.eta_e, 1.1f, 2.f);
			// should_update_materials |= overlay->sliderFloat(u8"##Absorption_oR", &uboParams.material.mu_o.r, .0f, 1.f); ImGui::SameLine();
			// should_update_materials |= overlay->sliderFloat(u8"##Absorption_oG", &uboParams.material.mu_o.g, .0f, 1.f); ImGui::SameLine();
			// should_update_materials |= overlay->sliderFloat(u8"μ_o##Absorption_oB", &uboParams.material.mu_o.b, .0f, 1.f);
			// should_update_materials |= overlay->sliderFloat(u8"##Absorption_eR", &uboParams.material.mu_e.r, .0f, 1.f); ImGui::SameLine();
			// should_update_materials |= overlay->sliderFloat(u8"##Absorption_eG", &uboParams.material.mu_e.g, .0f, 1.f); ImGui::SameLine();
			// should_update_materials |= overlay->sliderFloat(u8"μ_e##Absorption_eB", &uboParams.material.mu_e.b, .0f, 1.f);
		}

		if (overlay->header("Screenshot")) {
			static char screenshot_name[100] = {};
			// ImGui::InputText("##NameScreenshot", screenshot_name, 100, ImGuiInputTextFlags_AutoSelectAll, nullptr, nullptr);
			// ImGui::SameLine();
			if (overlay->button("Capture Screenshot")) {
				std::string name = screenshot_name;
				if (!name.length()) {
					auto t = std::time(nullptr);
					auto tm = *std::localtime(&t);

					std::ostringstream oss;
					oss << getAssetPath() << "../screenshot" << std::put_time(&tm, "%d-%m-%Y %H-%M-%S") << ".png";
					name = oss.str();
				}

				saveScreenshot(name.c_str());
			}
			
			static char capture_name[100] = {};
			// ImGui::InputText("##NameCapture", capture_name, 100, ImGuiInputTextFlags_AutoSelectAll, nullptr, nullptr);
			// ImGui::SameLine();
			if (overlay->comboBox("Capture diff image", &capture_mode, { std::string(""), 
																		 std::string("3-samples numeric"), 
																		 std::string("8-samples numeric"), 
																		 std::string("32-samples numeric"), 
																		 std::string("128-samples numeric"), 
																		 std::string("512-samples numeric"), 
																		 std::string("2048-samples numeric"), 
																		 std::string("Ground truth (16384-samples numeric)") })) {
				std::string name = capture_name;
				if (!name.length()) {
					auto t = std::time(nullptr);
					auto tm = *std::localtime(&t);

					std::ostringstream oss;
					oss << getAssetPath() << "../screenshot" << std::put_time(&tm, "%d-%m-%Y %H-%M-%S") << ".png";
					name = oss.str();
				}

				switch (capture_mode) {
					case 1:  beginDeltaCapture(name, 3, m_idx);     break;
					case 2:  beginDeltaCapture(name, 8, m_idx);     break;
					case 3:  beginDeltaCapture(name, 32, m_idx);    break;
					case 4:  beginDeltaCapture(name, 128, m_idx);   break;
					case 5:  beginDeltaCapture(name, 512, m_idx);   break;
					case 6:  beginDeltaCapture(name, 2048, m_idx);  break;
					case 7:  beginDeltaCapture(name, 16384, m_idx); break;
				}
				should_rebuild_pipelines = true;
			}
		}

		if (should_rebuild_pipelines)
		{
			// Need to rebuild pipeline object with new specialization constant. Wait for rendering to finish.
			vkDeviceWaitIdle(device);
			preparePipelines();
			reBuildCommandBuffers();
		}
		if (should_update_materials)
			updateParams();
	}
};

VULKAN_EXAMPLE_MAIN()

